@@ -29,8 +29,8 @@
 #include <sys/socket.h>
 
 #include <linux/socket.h>
-#include <linux/nfc.h>
 
+#include <near/nfc_copy.h>
 #include <near/plugin.h>
 #include <near/log.h>
 #include <near/types.h>
@@ -92,6 +92,22 @@
 
 #define MAX_DATA_SIZE	254
 
+#define NDEF_MAPPING_VERSION	0x10
+#define MAX_READ_BLOCKS_PER_CHECK	0x04
+#define MAX_WRITE_BLOCKS_PER_UPDATE	0x01
+#define MAX_BLOCKS_FOR_NDEF_DATA	0x000D
+#define ATTR_BLOCK_WRITE_FLAG	0x00
+#define ATTR_BLOCK_RW_FLAG	0x01
+
+#define IC_TYPE_OFFSET	12
+#define SYSTEM_OPTION_OFFSET	17
+#define FELICA_LITE_MC_BLOCK	0x88
+#define FELICA_LITE_IC_TYPE	0xF0
+#define FELICA_LITE_S_IC_TYPE	0xF1
+#define FELICA_PLUG_IC_TYPE	0xE0
+
+#define FELICA_LITE_AND_LITE_S_SYS_CODE	0x88B4
+
 struct type3_cmd {
 	uint8_t len;
 	uint8_t cmd;
@@ -115,12 +131,21 @@ struct t3_cookie {
 	uint8_t current_block;
 	uint8_t attr[BLOCK_SIZE];
 	struct near_ndef_message *ndef;
+	uint8_t ic_type;
+	uint8_t mc_block[BLOCK_SIZE];
 };
 
-static void t3_cookie_release(struct t3_cookie *cookie)
+static int t3_cookie_release(int err, void *data)
 {
+	struct t3_cookie *cookie = data;
+
+	DBG("%p", cookie);
+
 	if (cookie == NULL)
-		return;
+		return err;
+
+	if (cookie->cb != NULL)
+		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
 
 	if (cookie->ndef != NULL)
 		g_free(cookie->ndef->data);
@@ -128,9 +153,11 @@ static void t3_cookie_release(struct t3_cookie *cookie)
 	g_free(cookie->ndef);
 	g_free(cookie);
 	cookie = NULL;
+
+	return err;
 }
 
-/* common: Intialize structure to write block */
+/* common: Initialize structure to write block */
 static void prepare_write_block(uint8_t *UID, struct type3_cmd *cmd,
 					uint8_t block, uint8_t *data)
 {
@@ -147,15 +174,14 @@ static void prepare_write_block(uint8_t *UID, struct type3_cmd *cmd,
 	memcpy(cmd->data + LEN_ID + 6, data, BLOCK_SIZE); /* data to write */
 
 	cmd->len = LEN_ID + LEN_CMD + LEN_CMD_LEN + 6 + BLOCK_SIZE;
-
 }
 
 /* common: Initialize structure to read block */
 static void prepare_read_block(uint8_t cur_block,
 				uint8_t *UID,
-				struct type3_cmd *cmd )
+				struct type3_cmd *cmd)
 {
-	cmd->cmd	 = CMD_READ_WO_ENCRYPT;		/* command */
+	cmd->cmd = CMD_READ_WO_ENCRYPT;			/* command */
 	memcpy(cmd->data, UID, LEN_ID);			/* IDm */
 
 	cmd->data[LEN_ID] = 1;				/* number of service */
@@ -172,22 +198,20 @@ static void prepare_read_block(uint8_t cur_block,
 /* common: Simple checks on received frame */
 static int check_recv_frame(uint8_t *resp, uint8_t reply_code)
 {
-	int err = 0;
-
 	if (resp[OFS_NFC_STATUS] != 0) {
-		DBG("NFC Command failed: 0x%x",resp[OFS_NFC_STATUS]);
-		err = -EIO;
+		DBG("NFC Command failed: 0x%x", resp[OFS_NFC_STATUS]);
+		return -EIO;
 	}
 
-	if (resp[OFS_CMD_RESP] != reply_code ) {
+	if (resp[OFS_CMD_RESP] != reply_code) {
 		DBG("Felica cmd failed: 0x%x", resp[OFS_CMD_RESP]);
-		err = -EIO;
+		return -EIO;
 	}
 
-	return err;
+	return 0;
 }
 
-static int nfctype3_data_recv(uint8_t *resp, int length, void *data)
+static int data_recv(uint8_t *resp, int length, void *data)
 {
 	struct type3_tag *tag = data;
 	struct type3_cmd cmd;
@@ -205,11 +229,11 @@ static int nfctype3_data_recv(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	nfc_data = near_tag_get_data(tag->tag, &data_length);
-	length_read = length - OFS_READ_DATA  ;
+	length_read = length - OFS_READ_DATA;
 	current_length = tag->current_block * BLOCK_SIZE;
 	if (current_length + (length - OFS_READ_DATA) > data_length)
 		length_read = data_length - current_length;
@@ -222,7 +246,7 @@ static int nfctype3_data_recv(uint8_t *resp, int length, void *data)
 		tag->current_block = 0;
 
 		DBG("Done reading %zd bytes at %p", data_length, nfc_data);
-		records = near_ndef_parse(nfc_data, data_length);
+		records = near_ndef_parse_msg(nfc_data, data_length, NULL);
 		near_tag_add_records(tag->tag, records, tag->cb, 0);
 
 		g_free(tag);
@@ -235,16 +259,17 @@ static int nfctype3_data_recv(uint8_t *resp, int length, void *data)
 	tag->current_block += read_blocks;
 
 	prepare_read_block(DATA_BLOCK_START + tag->current_block,
-				tag->IDm, &cmd );
+				tag->IDm, &cmd);
+
+	err = near_adapter_send(adapter_idx, (uint8_t *) &cmd, cmd.len,
+				data_recv, tag, NULL);
 
-	err = near_adapter_send(adapter_idx, (uint8_t *)&cmd, cmd.len,
-			nfctype3_data_recv, tag);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	return 0;
 
-out:
+out_err:
 	if (err < 0 && tag->cb)
 		tag->cb(adapter_idx, target_idx, err);
 
@@ -253,7 +278,7 @@ out:
 	return err;
 }
 
-static int nfctype3_data_read(struct type3_tag *tag)
+static int data_read(struct type3_tag *tag)
 {
 	struct type3_cmd cmd;
 	uint32_t adapter_idx;
@@ -263,13 +288,13 @@ static int nfctype3_data_read(struct type3_tag *tag)
 	tag->current_block = 0;
 
 	prepare_read_block(DATA_BLOCK_START + tag->current_block,
-							tag->IDm, &cmd );
+							tag->IDm, &cmd);
 
 	adapter_idx = near_tag_get_adapter_idx(tag->tag);
 
 	return near_adapter_send(adapter_idx,
 					(uint8_t *) &cmd, cmd.len,
-					nfctype3_data_recv, tag);
+					data_recv, tag, NULL);
 }
 
 /* Read block 0 to retrieve the data length */
@@ -279,23 +304,23 @@ static int nfctype3_recv_block_0(uint8_t *resp, int length, void *data)
 	int err = 0;
 	struct near_tag *tag;
 	struct type3_tag *t3_tag = NULL;
-	uint32_t  ndef_data_length;
+	uint32_t ndef_data_length;
 
 	DBG("%d", length);
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	err = check_recv_frame(resp, RESP_READ_WO_ENCRYPT);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	if (resp[OFS_READ_FLAG] != 0) {
 		DBG("Status 0x%x", resp[OFS_READ_FLAG]);
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	/* Block 0:[11 - 13]: length is a 3 bytes value */
@@ -308,12 +333,12 @@ static int nfctype3_recv_block_0(uint8_t *resp, int length, void *data)
 	err = near_tag_add_data(cookie->adapter_idx, cookie->target_idx,
 					NULL, ndef_data_length);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	tag = near_tag_get_tag(cookie->adapter_idx, cookie->target_idx);
 	if (tag == NULL) {
 		err = -ENOMEM;
-		goto out;
+		goto out_err;
 	}
 
 	/* Block 0:[10]: RW Flag. 1 for RW */
@@ -325,69 +350,178 @@ static int nfctype3_recv_block_0(uint8_t *resp, int length, void *data)
 	t3_tag = g_try_malloc0(sizeof(struct type3_tag));
 	if (t3_tag == NULL) {
 		err = -ENOMEM;
-		goto out;
+		goto out_err;
 	}
 
-	memcpy(t3_tag->IDm, cookie->IDm , LEN_ID);
+	memcpy(t3_tag->IDm, cookie->IDm, LEN_ID);
 
 	near_tag_set_idm(tag, cookie->IDm, LEN_ID);
 	near_tag_set_attr_block(tag, resp + OFS_READ_DATA, BLOCK_SIZE);
+	near_tag_set_blank(tag, FALSE);
+	near_tag_set_ic_type(tag, cookie->ic_type);
 
 	t3_tag->adapter_idx = cookie->adapter_idx;
 	t3_tag->cb = cookie->cb;
 	t3_tag->tag = tag;
 
-	err = nfctype3_data_read(t3_tag);
+	err = data_read(t3_tag);
 
-out:
-	if (err < 0 && cookie->cb) {
-		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
+	/*
+	 * As reading isn't complete,
+	 * callback shouldn't be called while freeing the cookie
+	 */
+	if (err == 0)
+		cookie->cb = NULL;
+
+out_err:
+	if (err < 0)
 		g_free(t3_tag);
-	}
 
-	t3_cookie_release(cookie);
-
-	return err;
+	return t3_cookie_release(err, cookie);
 }
 
-static int nfctype3_recv_UID(uint8_t *resp, int length, void *data)
+static int poll_ndef_system_code(uint8_t *resp, int length, void *data)
 {
-	struct t3_cookie *rcv_cookie = data;
-	struct t3_cookie *snd_cookie;
+	struct t3_cookie *cookie = data;
 	int err = 0;
 	struct type3_cmd cmd;
 
-	DBG(" length: %d", length);
+	DBG("length: %d", length);
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	err = check_recv_frame(resp, RESP_POLL);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
-	snd_cookie = g_try_malloc0(sizeof(struct t3_cookie));
-	snd_cookie->adapter_idx = rcv_cookie->adapter_idx;
-	snd_cookie->target_idx = rcv_cookie->target_idx;
-	snd_cookie->cb = rcv_cookie->cb;
+	prepare_read_block(META_BLOCK_START, cookie->IDm, &cmd);
 
-	memcpy(snd_cookie->IDm, resp + OFS_IDM, LEN_ID);
+	err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
+				cmd.len, nfctype3_recv_block_0, cookie, NULL);
+	if (err < 0)
+		goto out_err;
 
-	prepare_read_block(META_BLOCK_START, snd_cookie->IDm, &cmd);
+	return 0;
 
-	err = near_adapter_send(snd_cookie->adapter_idx,
-			(uint8_t *)&cmd, cmd.len, nfctype3_recv_block_0, snd_cookie);
+out_err:
+	return t3_cookie_release(err, cookie);
+}
 
-out:
-	if (err < 0 && rcv_cookie->cb)
-		rcv_cookie->cb(rcv_cookie->adapter_idx,
-				rcv_cookie->target_idx, err);
+static int check_sys_op_in_mc_block(uint8_t *resp, int length, void *data)
+{
+	struct type3_cmd cmd;
+	struct near_tag *tag;
+	struct t3_cookie *cookie = data;
+	int err = 0;
 
-	t3_cookie_release(rcv_cookie);
+	DBG("length %d", length);
 
-	return err;
+	if (length < 0) {
+		err = -EIO;
+		goto out_err;
+	}
+
+	err = check_recv_frame(resp, RESP_READ_WO_ENCRYPT);
+	if (err < 0)
+		goto out_err;
+
+	if (resp[SYSTEM_OPTION_OFFSET] == 0x00) {
+		DBG("Blank tag detected");
+
+		err = near_tag_add_data(cookie->adapter_idx,
+					cookie->target_idx,
+					NULL, 1 /* dummy length */);
+		if (err < 0)
+			goto out_err;
+
+		tag = near_tag_get_tag(cookie->adapter_idx,
+					cookie->target_idx);
+		if (tag == NULL) {
+			err = -ENOMEM;
+		goto out_err;
+		}
+
+		near_tag_set_idm(tag, cookie->IDm, LEN_ID);
+		near_tag_set_ic_type(tag, cookie->ic_type);
+		near_tag_set_blank(tag, TRUE);
+
+		return t3_cookie_release(0, cookie);
+	} else {
+		/* CMD POLL */
+		cmd.cmd	 = CMD_POLL;	/* POLL command */
+		cmd.data[0] = 0x12;     /* System code (NFC SC) */
+		cmd.data[1] = 0xFC;
+		cmd.data[2] = 01;	/* request code */
+		cmd.data[3] = 0x00;	/* time slot */
+		/* data len + 2 bytes */
+		cmd.len = LEN_CMD + LEN_CMD_LEN + 4 ;
+
+		err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
+			cmd.len , poll_ndef_system_code, cookie, NULL);
+	}
+
+	if (err < 0)
+		goto out_err;
+
+	return 0;
+
+out_err:
+	return t3_cookie_release(err, cookie);
+}
+
+static int receive_system_code(uint8_t *resp, int length, void *data)
+{
+	struct t3_cookie *cookie = data;
+	int err = 0;
+	struct type3_cmd cmd;
+
+	DBG("length: %d", length);
+
+	if (length < 0) {
+		err = -EIO;
+		goto out_err;
+	}
+
+	err = check_recv_frame(resp, RESP_POLL);
+	if (err < 0)
+		goto out_err;
+
+	cookie->ic_type = resp[IC_TYPE_OFFSET];
+	memcpy(cookie->IDm, resp + OFS_IDM, LEN_ID);
+
+	switch (resp[IC_TYPE_OFFSET]) {
+	case FELICA_LITE_IC_TYPE:
+		prepare_read_block(FELICA_LITE_MC_BLOCK, cookie->IDm, &cmd);
+		err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
+					cmd.len, check_sys_op_in_mc_block,
+					cookie, NULL);
+		break;
+
+	default:
+		/* CMD POLL */
+		cmd.cmd	 = CMD_POLL;	/* POLL command */
+		cmd.data[0] = 0x12;     /* System code (NFC SC) */
+		cmd.data[1] = 0xFC;
+		cmd.data[2] = 01;	/* request code */
+		cmd.data[3] = 0x00;	/* time slot */
+		/* data len + 2 bytes */
+		cmd.len = LEN_CMD + LEN_CMD_LEN + 4 ;
+
+		err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
+					cmd.len, poll_ndef_system_code, cookie,
+					NULL);
+	}
+
+	if (err < 0)
+		goto out_err;
+
+	return 0;
+
+out_err:
+	return t3_cookie_release(err, cookie);
 }
 
 static int nfctype3_read(uint32_t adapter_idx,
@@ -400,21 +534,25 @@ static int nfctype3_read(uint32_t adapter_idx,
 
 	/* CMD POLL */
 	cmd.cmd	 = CMD_POLL;	/* POLL command */
-	cmd.data[0] = 0x12;     /* System code (NFC SC) */
-	cmd.data[1] = 0xFC;
+	cmd.data[0] = 0xFF;     /* System code */
+	cmd.data[1] = 0xFF;
 	cmd.data[2] = 01;	/* request code */
 	cmd.data[3] = 0x00;	/* time slot */
 
 	/* data len + 2 bytes */
-	cmd.len = LEN_CMD + LEN_CMD_LEN + 4 ;
+	cmd.len = LEN_CMD + LEN_CMD_LEN + 4;
 
 	cookie = g_try_malloc0(sizeof(struct t3_cookie));
+	if (cookie == NULL)
+		return -ENOMEM;
+
 	cookie->adapter_idx = adapter_idx;
 	cookie->target_idx = target_idx;
 	cookie->cb = cb;
 
-	return near_adapter_send(adapter_idx, (uint8_t *)&cmd,
-			cmd.len , nfctype3_recv_UID, cookie);
+	return near_adapter_send(adapter_idx, (uint8_t *) &cmd, cmd.len,
+					receive_system_code, cookie,
+					t3_cookie_release);
 }
 
 static int update_attr_block_cb(uint8_t *resp, int length, void *data)
@@ -426,22 +564,17 @@ static int update_attr_block_cb(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	err = check_recv_frame(resp, RESP_WRITE_WO_ENCRYPT);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	DBG("Done writing");
 
-out:
-	if (cookie->cb)
-		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
-
-	t3_cookie_release(cookie);
-
-	return err;
+out_err:
+	return t3_cookie_release(err, cookie);
 }
 
 static int update_attr_block(struct t3_cookie *cookie)
@@ -467,7 +600,8 @@ static int update_attr_block(struct t3_cookie *cookie)
 	prepare_write_block(cookie->IDm, &cmd, 0, cookie->attr);
 
 	return near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd, cmd.len,
-						update_attr_block_cb, cookie);
+					update_attr_block_cb, cookie,
+					t3_cookie_release);
 }
 
 static int data_write_resp(uint8_t *resp, int length, void *data)
@@ -481,17 +615,17 @@ static int data_write_resp(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	err = check_recv_frame(resp, RESP_WRITE_WO_ENCRYPT);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	if (cookie->ndef->offset >= cookie->ndef->length) {
 		err = update_attr_block(cookie);
 		if (err < 0)
-			goto out;
+			goto out_err;
 
 		return 0;
 	}
@@ -510,20 +644,15 @@ static int data_write_resp(uint8_t *resp, int length, void *data)
 	cookie->current_block++;
 	cookie->ndef->offset += BLOCK_SIZE;
 
-	err =  near_adapter_send(cookie->adapter_idx, (uint8_t *)&cmd, cmd.len,
-						data_write_resp, cookie);
+	err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd, cmd.len,
+						data_write_resp, cookie, NULL);
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	return 0;
 
-out:
-	if (err < 0 && cookie->cb)
-		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
-
-	t3_cookie_release(cookie);
-
-	return err;
+out_err:
+	return t3_cookie_release(err, cookie);
 }
 
 static int data_write(uint32_t adapter_idx, uint32_t target_idx,
@@ -541,9 +670,14 @@ static int data_write(uint32_t adapter_idx, uint32_t target_idx,
 	DBG("");
 
 	cookie = g_try_malloc0(sizeof(struct t3_cookie));
+
 	if (cookie == NULL) {
 		err = -ENOMEM;
-		goto out;
+
+		if (cb != NULL)
+			cb(adapter_idx, target_idx, err);
+
+		return err;
 	}
 
 	cookie->adapter_idx = adapter_idx;
@@ -553,26 +687,22 @@ static int data_write(uint32_t adapter_idx, uint32_t target_idx,
 	cookie->current_block = 0;
 
 	idm = near_tag_get_idm(tag, &len);
-	if (idm == NULL) {
-		err = -EINVAL;
-		goto out;
-	}
+	if (idm == NULL)
+		return t3_cookie_release(-EINVAL, cookie);
 
 	memcpy(cookie->IDm, idm, len);
 
 	attr = near_tag_get_attr_block(tag, &len);
-	if (attr == NULL) {
-		err = -EINVAL;
-		goto out;
-	}
+	if (attr == NULL)
+		return t3_cookie_release(-EINVAL, cookie);
 
 	memcpy(cookie->attr, attr, len);
-	nmaxb = (((uint16_t)(cookie->attr[3])) << 8) | cookie->attr[4];
+	nmaxb = (((uint16_t) (cookie->attr[3])) << 8) | cookie->attr[4];
 
 	if (cookie->ndef->length > (nmaxb * BLOCK_SIZE)) {
 		near_error("not enough space on tag");
-		err = -ENOSPC;
-		goto out;
+
+		return t3_cookie_release(-ENOSPC, cookie);
 	}
 
 	cookie->attr[9] = 0x0F; /* writing data in progress */
@@ -581,24 +711,16 @@ static int data_write(uint32_t adapter_idx, uint32_t target_idx,
 	for (i = 0; i < 14; i++)
 		checksum += cookie->attr[i];
 
-	cookie->attr[14] = (uint8_t)(checksum >> 8);
-	cookie->attr[15] = (uint8_t)(checksum);
+	cookie->attr[14] = (uint8_t) (checksum >> 8);
+	cookie->attr[15] = (uint8_t) checksum;
 
 	prepare_write_block(cookie->IDm, &cmd, cookie->current_block,
 							cookie->attr);
 	cookie->current_block++;
 
-	err = near_adapter_send(adapter_idx, (uint8_t *)&cmd, cmd.len,
-						data_write_resp, cookie);
-	if (err < 0)
-		goto out;
-
-	return 0;
-
-out:
-	t3_cookie_release(cookie);
-
-	return err;
+	return near_adapter_send(adapter_idx, (uint8_t *) &cmd, cmd.len,
+					data_write_resp, cookie,
+					t3_cookie_release);
 }
 
 static int nfctype3_write(uint32_t adapter_idx, uint32_t target_idx,
@@ -606,22 +728,33 @@ static int nfctype3_write(uint32_t adapter_idx, uint32_t target_idx,
 				near_tag_io_cb cb)
 {
 	struct near_tag *tag;
+	int err;
 
 	DBG("");
 
-	if (ndef == NULL || cb == NULL)
-		return -EINVAL;
+	if (ndef == NULL || cb == NULL) {
+		err = -EINVAL;
+		goto out_err;
+	}
 
 	tag = near_tag_get_tag(adapter_idx, target_idx);
-	if (tag == NULL)
-		return -EINVAL;
+	if (tag == NULL) {
+		err = -EINVAL;
+		goto out_err;
+	}
 
-	return data_write(adapter_idx, target_idx, ndef, tag, cb);
+	err = data_write(adapter_idx, target_idx, ndef, tag, cb);
+
+out_err:
+	if (cb != NULL && err < 0)
+		cb(adapter_idx, target_idx, err);
+
+	return err;
 }
 
-static int nfctype3_check_recv_UID(uint8_t *resp, int length, void *data)
+static int check_presence(uint8_t *resp, int length, void *data)
 {
-	struct t3_cookie *rcv_cookie = data;
+	struct t3_cookie *cookie = data;
 	int err = 0;
 
 	DBG("length %d", length);
@@ -629,13 +762,7 @@ static int nfctype3_check_recv_UID(uint8_t *resp, int length, void *data)
 	if (length < 0)
 		err = -EIO;
 
-	if (rcv_cookie->cb)
-		rcv_cookie->cb(rcv_cookie->adapter_idx,
-				rcv_cookie->target_idx, err);
-
-	t3_cookie_release(rcv_cookie);
-
-	return err;
+	return t3_cookie_release(err, cookie);
 }
 
 static int nfctype3_check_presence(uint32_t adapter_idx,
@@ -643,14 +770,13 @@ static int nfctype3_check_presence(uint32_t adapter_idx,
 {
 	struct type3_cmd cmd;
 	struct t3_cookie *cookie;
-	int err;
 
 	DBG("");
 
 	/* CMD POLL */
 	cmd.cmd	 = CMD_POLL;	/* POLL command */
-	cmd.data[0] = 0x12;     /* System code (NFC SC) */
-	cmd.data[1] = 0xFC;
+	cmd.data[0] = 0xFF;     /* System code */
+	cmd.data[1] = 0xFF;
 	cmd.data[2] = 01;	/* request code */
 	cmd.data[3] = 0x00;	/* time slot */
 
@@ -665,17 +791,180 @@ static int nfctype3_check_presence(uint32_t adapter_idx,
 	cookie->target_idx = target_idx;
 	cookie->cb = cb;
 
-	err = near_adapter_send(adapter_idx, (uint8_t *)&cmd,
-			cmd.len , nfctype3_check_recv_UID, cookie);
+	return near_adapter_send(adapter_idx, (uint8_t *) &cmd,
+				cmd.len, check_presence, cookie,
+				t3_cookie_release);
+}
+
+static int format_resp(uint8_t *resp, int length, void *data)
+{
+	struct near_tag *tag;
+	struct t3_cookie *cookie = data;
+	int err;
+
+	DBG("");
+
+	if (length < 0) {
+		err = -EIO;
+		goto out_err;
+	}
+
+	err = check_recv_frame(resp, RESP_WRITE_WO_ENCRYPT);
 	if (err < 0)
-		goto out;
+		goto out_err;
+
+	tag = near_tag_get_tag(cookie->adapter_idx, cookie->target_idx);
+	if (tag == NULL) {
+		err = -ENOMEM;
+		goto out_err;
+	}
+
+	near_tag_set_ro(tag, FALSE);
+	near_tag_set_idm(tag, cookie->IDm, LEN_ID);
+	near_tag_set_attr_block(tag, cookie->attr, BLOCK_SIZE);
+	near_tag_set_blank(tag, FALSE);
+
+	DBG("Formatting is done");
+
+out_err:
+	return t3_cookie_release(err, cookie);
+}
+
+static int write_attr_block(uint8_t *resp, int length , void *data)
+{
+	struct type3_cmd cmd;
+	struct t3_cookie *cookie = data;
+	int err, i;
+	uint16_t checksum = 0;
+
+	DBG("length %d", length);
+
+	if (length < 0) {
+		err = -EIO;
+		goto out_err;
+	}
+
+	err = check_recv_frame(resp, RESP_WRITE_WO_ENCRYPT);
+	if (err < 0)
+		goto out_err;
+
+	cookie->attr[0] = NDEF_MAPPING_VERSION;
+	cookie->attr[1] = MAX_READ_BLOCKS_PER_CHECK;
+	cookie->attr[2] = MAX_WRITE_BLOCKS_PER_UPDATE;
+	cookie->attr[3] = (uint8_t) (MAX_BLOCKS_FOR_NDEF_DATA >> 8);
+	cookie->attr[4] = (uint8_t) (MAX_BLOCKS_FOR_NDEF_DATA);
+	cookie->attr[5] = 0;
+	cookie->attr[6] = 0;
+	cookie->attr[7] = 0;
+	cookie->attr[8] = 0;
+	cookie->attr[9] = ATTR_BLOCK_WRITE_FLAG;
+	cookie->attr[10] = ATTR_BLOCK_RW_FLAG;
+	cookie->attr[11] = 0;
+	cookie->attr[12] = 0;
+	cookie->attr[13] = 0;
+
+	for (i = 0; i < (BLOCK_SIZE - CHECKSUM_LEN); i++)
+		checksum += cookie->attr[i];
+
+	cookie->attr[14] = (uint8_t) (checksum >> 8);
+	cookie->attr[15] = (uint8_t) checksum;
+
+	prepare_write_block(cookie->IDm, &cmd, META_BLOCK_START,
+						cookie->attr);
+
+	err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
+				cmd.len, format_resp, cookie, NULL);
+	if (err < 0)
+		goto out_err;
 
 	return 0;
 
-out:
-	t3_cookie_release(cookie);
+out_err:
+	return t3_cookie_release(err, cookie);
+}
+
+static int write_mc_block(uint8_t *resp, int length, void *data)
+{
+	struct type3_cmd cmd;
+	struct t3_cookie *cookie = data;
+	int err;
+
+	DBG("length %d", length);
+
+	if (length < 0) {
+		err = -EIO;
+		goto out_err;
+	}
+
+	err = check_recv_frame(resp, RESP_READ_WO_ENCRYPT);
+	if (err < 0)
+		goto out_err;
+
+	if (resp[OFS_READ_FLAG] != 0) {
+		DBG("Status 0x%x", resp[OFS_READ_FLAG]);
+		err = -EIO;
+		goto out_err;
+	}
+
+	memcpy(cookie->mc_block, resp + 14, BLOCK_SIZE);
+	/*
+	 * By updating Byte3 to 01h means making Felica Lite
+	 * compatible with NDEF.
+	 */
+	cookie->mc_block[3] = 1;
+	prepare_write_block(cookie->IDm, &cmd, FELICA_LITE_MC_BLOCK,
+				cookie->mc_block);
+	err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
+				cmd.len, write_attr_block, cookie, NULL);
+	if (err < 0)
+		goto out_err;
+
+	return 0;
+
+out_err:
+	return t3_cookie_release(err, cookie);
+}
+
+static int nfctype3_format(uint32_t adapter_idx,
+				uint32_t target_idx, near_tag_io_cb cb)
+{
+	struct type3_cmd cmd;
+	struct near_tag *tag;
+	struct t3_cookie *cookie;
+	uint8_t ic_type;
+	uint8_t *idm, len;
+
+	DBG("");
+
+	tag = near_tag_get_tag(adapter_idx, target_idx);
+	if (tag == NULL)
+		return -ENOMEM;
+
+	ic_type = near_tag_get_ic_type(tag);
+	if (ic_type != FELICA_LITE_IC_TYPE)
+		return -EOPNOTSUPP;
+
+	cookie = g_try_malloc0(sizeof(struct t3_cookie));
+	if (cookie == NULL)
+		return -ENOMEM;
+
+	cookie->adapter_idx = adapter_idx;
+	cookie->target_idx = target_idx;
+	cookie->cb = cb;
+	cookie->ic_type = ic_type;
+
+	idm = near_tag_get_idm(tag, &len);
+	if (idm == NULL)
+		return t3_cookie_release(-EINVAL, cookie);
+
+	memcpy(cookie->IDm, idm, len);
+
+	prepare_read_block(FELICA_LITE_MC_BLOCK, cookie->IDm, &cmd);
+
+	return near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd, cmd.len,
+					write_mc_block, cookie,
+					t3_cookie_release);
 
-	return err;
 }
 
 static struct near_tag_driver type1_driver = {
@@ -683,6 +972,7 @@ static struct near_tag_driver type1_driver = {
 	.priority       = NEAR_TAG_PRIORITY_DEFAULT,
 	.read           = nfctype3_read,
 	.write          = nfctype3_write,
+	.format		= nfctype3_format,
 	.check_presence = nfctype3_check_presence,
 };
 
@@ -702,4 +992,3 @@ static void nfctype3_exit(void)
 
 NEAR_PLUGIN_DEFINE(nfctype3, "NFC Forum Type 3 tags support", VERSION,
 			NEAR_PLUGIN_PRIORITY_HIGH, nfctype3_init, nfctype3_exit)
-
