@@ -30,8 +30,8 @@
 #include <sys/socket.h>
 
 #include <linux/socket.h>
-#include <linux/nfc.h>
 
+#include <near/nfc_copy.h>
 #include <near/plugin.h>
 #include <near/log.h>
 #include <near/types.h>
@@ -53,10 +53,67 @@ struct p2p_data {
 	near_device_io_cb cb;
 	int fd;
 	guint watch;
+	struct p2p_data *server;
 
 	GList *client_list;
 };
 
+/* common function for socket binding */
+static int __p2p_bind(struct p2p_data *server_data, GIOFunc listener)
+{
+	int err, fd = server_data->fd;
+	struct sockaddr_nfc_llcp addr;
+	GIOChannel *channel;
+	struct near_p2p_driver *driver = server_data->driver;
+
+	DBG("Binding %s", driver->name);
+
+	memset(&addr, 0, sizeof(struct sockaddr_nfc_llcp));
+	addr.sa_family = AF_NFC;
+	addr.dev_idx = server_data->adapter_idx;
+	addr.nfc_protocol = NFC_PROTO_NFC_DEP;
+	addr.service_name_len = strlen(driver->service_name);
+	strcpy(addr.service_name, driver->service_name);
+
+	err = bind(fd, (struct sockaddr *) &addr,
+			sizeof(struct sockaddr_nfc_llcp));
+	if (err < 0) {
+		if (errno == EADDRINUSE) {
+			DBG("%s is already bound", driver->name);
+			err = 0;
+			goto out_err;
+		}
+
+		near_error("%s bind failed %d %d", driver->name, err, errno);
+		goto out_err;
+	}
+
+	if (server_data->driver->sock_type == SOCK_STREAM) {
+		err = listen(fd, 10);
+		if (err < 0) {
+			near_error("%s listen failed %d", driver->name, err);
+			goto out_err;
+		}
+	}
+
+	channel = g_io_channel_unix_new(fd);
+	g_io_channel_set_close_on_unref(channel, TRUE);
+
+	server_data->watch = g_io_add_watch(channel,
+				G_IO_IN | G_IO_HUP | G_IO_NVAL | G_IO_ERR,
+				listener, (gpointer) server_data);
+	g_io_channel_unref(channel);
+
+	return 0;
+
+out_err:
+	close(fd);
+	return err;
+}
+
+static gboolean p2p_listener_event(GIOChannel *channel, GIOCondition condition,
+				   gpointer user_data);
+
 static gboolean p2p_client_event(GIOChannel *channel, GIOCondition condition,
 							gpointer user_data)
 {
@@ -67,6 +124,7 @@ static gboolean p2p_client_event(GIOChannel *channel, GIOCondition condition,
 
 	if (condition & (G_IO_NVAL | G_IO_ERR | G_IO_HUP)) {
 		int err;
+		struct p2p_data *server_data = client_data->server;
 
 		if (client_data->watch > 0)
 			g_source_remove(client_data->watch);
@@ -83,6 +141,21 @@ static gboolean p2p_client_event(GIOChannel *channel, GIOCondition condition,
 		near_error("%s client channel closed",
 					client_data->driver->name);
 
+		if (server_data)
+			server_data->client_list =
+				g_list_remove(server_data->client_list,
+								client_data);
+
+		if (client_data->driver->single_connection) {
+			server_data->fd = socket(AF_NFC, SOCK_STREAM,
+							NFC_SOCKPROTO_LLCP);
+			if (server_data->fd > 0)
+				__p2p_bind(server_data, p2p_listener_event);
+		}
+
+		if (client_data->driver->sock_type == SOCK_STREAM)
+			g_free(client_data);
+
 		return FALSE;
 	}
 
@@ -91,12 +164,6 @@ static gboolean p2p_client_event(GIOChannel *channel, GIOCondition condition,
 						client_data->target_idx,
 						client_data->cb);
 
-	if (more == FALSE) {
-		if (client_data->driver->close != NULL)
-			client_data->driver->close(client_data->fd, 0);
-		close(client_data->fd);
-	}
-
 	return more;
 }
 
@@ -106,7 +173,7 @@ static void free_client_data(gpointer data)
 
 	DBG("");
 
-	client_data = (struct p2p_data *)data;
+	client_data = (struct p2p_data *) data;
 
 	if (client_data->driver->close != NULL)
 		client_data->driver->close(client_data->fd, 0);
@@ -129,10 +196,13 @@ static void free_server_data(gpointer data)
 		g_source_remove(server_data->watch);
 	server_data->watch = 0;
 	g_list_free_full(server_data->client_list, free_client_data);
+	server_data->client_list = NULL;
 
 	DBG("Closing server socket");
 
 	close(server_data->fd);
+
+	g_free(server_data);
 }
 
 static gboolean p2p_listener_event(GIOChannel *channel, GIOCondition condition,
@@ -153,9 +223,9 @@ static gboolean p2p_listener_event(GIOChannel *channel, GIOCondition condition,
 		if (server_data->watch > 0)
 			g_source_remove(server_data->watch);
 		server_data->watch = 0;
-		g_list_free_full(server_data->client_list, free_client_data);
 
-		close(server_fd);
+		g_list_free_full(server_data->client_list, free_client_data);
+		server_data->client_list = NULL;
 
 		near_error("Error with %s server channel", driver->name);
 
@@ -163,12 +233,11 @@ static gboolean p2p_listener_event(GIOChannel *channel, GIOCondition condition,
 	}
 
 	client_addr_len = sizeof(client_addr);
-	client_fd = accept(server_fd, (struct sockaddr *)&client_addr,
+	client_fd = accept(server_fd, (struct sockaddr *) &client_addr,
 							&client_addr_len);
 	if (client_fd < 0) {
 		near_error("accept failed %d", client_fd);
 
-		close(server_fd);
 		return FALSE;
 	}
 
@@ -187,6 +256,7 @@ static gboolean p2p_listener_event(GIOChannel *channel, GIOCondition condition,
 	client_data->target_idx = client_addr.target_idx;
 	client_data->fd = client_fd;
 	client_data->cb = server_data->cb;
+	client_data->server = server_data;
 
 	client_channel = g_io_channel_unix_new(client_fd);
 	g_io_channel_set_close_on_unref(client_channel, TRUE);
@@ -196,54 +266,37 @@ static gboolean p2p_listener_event(GIOChannel *channel, GIOCondition condition,
 				p2p_client_event,
 				client_data);
 
+	g_io_channel_unref(client_channel);
+
 	server_data->client_list = g_list_append(server_data->client_list, client_data);
 
-	return TRUE;
+	return !driver->single_connection;
 }
 
 static int p2p_bind(struct near_p2p_driver *driver, uint32_t adapter_idx,
 							near_device_io_cb cb)
 {
 	int err, fd;
-	struct sockaddr_nfc_llcp addr;
-	GIOChannel *channel;
 	struct p2p_data *server_data;
+	GIOFunc g_func;
 
-	DBG("Binding %s", driver->name);
+	DBG("");
 
-	fd = socket(AF_NFC, SOCK_STREAM, NFC_SOCKPROTO_LLCP);
+	if (driver->sock_type == SOCK_DGRAM)
+		g_func = p2p_client_event;
+	else
+		g_func = p2p_listener_event;
+
+	if (driver->sock_type != SOCK_DGRAM &&
+					driver->sock_type != SOCK_STREAM) {
+		near_error("Undefined socket type for %s ", driver->name);
+		return -EINVAL;
+	}
+
+	fd = socket(AF_NFC, driver->sock_type, NFC_SOCKPROTO_LLCP);
 	if (fd < 0)
 		return -errno;
 
-	memset(&addr, 0, sizeof(struct sockaddr_nfc_llcp));
-	addr.sa_family = AF_NFC;
-	addr.dev_idx = adapter_idx;
-	addr.nfc_protocol = NFC_PROTO_NFC_DEP;
-	addr.service_name_len = strlen(driver->service_name);
-	strcpy(addr.service_name, driver->service_name);
-
-	err = bind(fd, (struct sockaddr *)&addr,
-			sizeof(struct sockaddr_nfc_llcp));
-	if (err < 0) {
-		if (errno == EADDRINUSE) {
-			DBG("%s is already bound", driver->name);
-			return 0;
-		}
-
-		near_error("%s bind failed %d %d", driver->name, err, errno);
-
-		close(fd);
-		return err;
-	}
-
-	err = listen(fd, 10);
-	if (err < 0) {
-		near_error("%s listen failed %d", driver->name, err);
-
-		close(fd);
-		return err;
-	}
-
 	server_data = g_try_malloc0(sizeof(struct p2p_data));
 	if (server_data == NULL) {
 		close(fd);
@@ -255,13 +308,11 @@ static int p2p_bind(struct near_p2p_driver *driver, uint32_t adapter_idx,
 	server_data->fd = fd;
 	server_data->cb = cb;
 
-	channel = g_io_channel_unix_new(fd);
-	g_io_channel_set_close_on_unref(channel, TRUE);
-
-	server_data->watch = g_io_add_watch(channel,
-				G_IO_IN | G_IO_HUP | G_IO_NVAL | G_IO_ERR,
-				p2p_listener_event,
-				(gpointer) server_data);
+	err = __p2p_bind(server_data, g_func);
+	if (err < 0) {
+		g_free(server_data);
+		return err;
+	}
 
 	server_list = g_list_append(server_list, server_data);
 
@@ -270,13 +321,19 @@ static int p2p_bind(struct near_p2p_driver *driver, uint32_t adapter_idx,
 
 static int p2p_listen(uint32_t adapter_idx, near_device_io_cb cb)
 {
-	int err = 0;
+	int err = -1, bind_err;
 	GSList *list;
 
 	for (list = driver_list; list != NULL; list = list->next) {
 		struct near_p2p_driver *driver = list->data;
 
-		err &= p2p_bind(driver, adapter_idx, cb);
+		bind_err = p2p_bind(driver, adapter_idx, cb);
+		if (bind_err == 0) {
+			err = 0;
+		} else if (bind_err == -EPROTONOSUPPORT) {
+			near_error("LLCP is not supported");
+			return bind_err;
+		}
 	}
 
 	return err;
@@ -303,7 +360,7 @@ static int p2p_connect(uint32_t adapter_idx, uint32_t target_idx,
 	addr.service_name_len = strlen(driver->service_name);
 	strcpy(addr.service_name, driver->service_name);
 
-	err = connect(fd, (struct sockaddr *)&addr,
+	err = connect(fd, (struct sockaddr *) &addr,
 			sizeof(struct sockaddr_nfc_llcp));
 	if (err < 0) {
 		near_error("Connect failed  %d", err);
@@ -329,12 +386,24 @@ static int p2p_push(uint32_t adapter_idx, uint32_t target_idx,
 
 		if (strcmp(driver->service_name, service_name) != 0)
 			continue;
-
+		/*
+		 * Because of Android's implementation, we have use SNEP for
+		 * Handover. So, on Handover session, we try to connect to
+		 * the handover service and fallback to SNEP on connect fail.
+		 */
 		fd = p2p_connect(adapter_idx, target_idx, ndef, cb, driver);
-		if (fd < 0)
-			return fd;
+		if (fd < 0) {
+			if (driver->fallback_service_name != NULL)
+				return  p2p_push(adapter_idx, target_idx, ndef,
+					(char *) driver->fallback_service_name,
+					cb);
+			else
+				return -1;
 
-		return driver->push(fd, adapter_idx, target_idx, ndef, cb);
+		} else if (driver->push != NULL) {
+			return driver->push(fd, adapter_idx, target_idx,
+					ndef, cb);
+		}
 	}
 
 	return -1;
@@ -369,6 +438,8 @@ static int p2p_init(void)
 
 	npp_init();
 	snep_init();
+	snep_validation_init();
+	llcp_validation_init();
 	handover_init();
 
 	return near_device_driver_register(&p2p_driver);
@@ -380,7 +451,9 @@ static void p2p_exit(void)
 
 	g_list_free_full(server_list, free_server_data);
 
+	llcp_validation_exit();
 	snep_exit();
+	snep_validation_exit();
 	npp_exit();
 	handover_exit();
 
