@@ -26,6 +26,9 @@
 struct near_p2p_driver {
 	const char *name;
 	const char *service_name;
+	const char *fallback_service_name;
+	near_bool_t single_connection;
+	int sock_type;
 	near_bool_t (*read)(int client_fd,
 				uint32_t adapter_idx, uint32_t target_idx,
 				near_device_io_cb cb);
@@ -34,16 +37,20 @@ struct near_p2p_driver {
 	void (*close)(int client_fd, int err);
 };
 
-#define TLV_SIZE 2
-
 int npp_init(void);
 void npp_exit(void);
 
-int snep_init(void);
-void snep_exit(void);
-
 int handover_init(void);
 void handover_exit(void);
 
+int snep_init(void);
+void snep_exit(void);
+
+int snep_validation_init(void);
+void snep_validation_exit(void);
+
+int llcp_validation_init(void);
+void llcp_validation_exit(void);
+
 int near_p2p_register(struct near_p2p_driver *driver);
 void near_p2p_unregister(struct near_p2p_driver *driver);
