@@ -597,6 +597,7 @@ struct near_adapter *__near_adapter_create(uint32_t idx,
 		const char *name, uint32_t protocols, near_bool_t powered)
 {
 	struct near_adapter *adapter;
+	near_bool_t powered_setting;
 
 	adapter = g_try_malloc0(sizeof(struct near_adapter));
 	if (adapter == NULL)
@@ -607,6 +608,14 @@ struct near_adapter *__near_adapter_create(uint32_t idx,
 		g_free(adapter);
 		return NULL;
 	}
+
+	powered_setting = near_setting_get_bool("DefaultPowered");
+	if (powered_setting == TRUE && powered == FALSE &&
+	    !__near_netlink_adapter_enable(adapter->idx, powered_setting))
+			powered = TRUE;
+
+	DBG("Powered %d", powered);
+
 	adapter->idx = idx;
 	adapter->protocols = protocols;
 	adapter->powered = powered;
@@ -789,7 +798,15 @@ static int adapter_add_tag(struct near_adapter *adapter, uint32_t target_idx,
 		return err;
 	}
 
-	return __near_tag_read(tag, tag_read_cb);
+	err = __near_tag_read(tag, tag_read_cb);
+	if (err < 0) {
+		near_error("Could not read the tag");
+
+		near_adapter_disconnect(adapter->idx);
+		__near_adapter_remove_target(adapter->idx, target_idx);
+	}
+
+	return err;
 }
 
 static int adapter_add_device(struct near_adapter *adapter,
