@@ -38,12 +38,16 @@
 #define DEFAULT_ADAPTER			"DefaultAdapter"
 #define ADAPTER_REMOVED			"AdapterRemoved"
 #define DEFAULT_ADAPTER_CHANGED		"DefaultAdapterChanged"
+#define ADAPTER_PROPERTY_CHANGED	"PropertyChanged"
 #define MANAGER_PATH			"/"
 #define OOB_AGENT			"/org/neard/agent/neard_oob"
 
 #define BT_NOINPUTOUTPUT		"NoInputNoOutput"
 #define BT_DISPLAY_YESNO		"DisplayYesNo"
 
+#define DBUS_MANAGER_INTF		"org.freedesktop.DBus.ObjectManager"
+#define AGENT_REGISTER_TIMEOUT	2
+
 /* BT EIR list */
 #define EIR_UUID128_ALL		0x07 /* 128-bit UUID, all listed */
 #define EIR_NAME_SHORT		0x08 /* shortened local name */
@@ -62,6 +66,7 @@
 #define BT_ADDRESS_SIZE		6
 #define COD_SIZE		3
 #define OOB_SP_SIZE		16
+#define EIR_SIZE_MAX		255
 
 struct near_oob_data {
 	char *def_adapter;
@@ -87,7 +92,12 @@ struct near_oob_data {
 static DBusConnection *bt_conn;
 static struct near_oob_data bt_def_oob_data;
 
-static int bt_do_pairing(struct near_oob_data *oob);
+static guint watch;
+static guint removed_watch;
+static guint adapter_watch;
+static guint adapter_props_watch;
+
+static guint register_bluez_timer;
 
 static void __bt_eir_free(struct near_oob_data *oob)
 {
@@ -117,8 +127,6 @@ static void __bt_eir_free(struct near_oob_data *oob)
 		g_free(oob->spair_randomizer);
 		oob->spair_randomizer = NULL;
 	}
-
-
 }
 
 static void bt_eir_free(struct near_oob_data *oob)
@@ -149,7 +157,7 @@ static int bt_generic_call(DBusConnection *conn,
 
 	if (msg == NULL) {
 		near_error("Unable to allocate new D-Bus %s message", method);
-		err = -ENOMEM;
+		return -ENOMEM;
 	}
 
 	va_start(args, type);
@@ -203,7 +211,7 @@ static void bt_create_paired_device_cb(DBusPendingCall *pending,
 		goto cb_done;
 	}
 
-	near_info("Pairing done successfully !");
+	DBG("Successful pairing");
 
 cb_done:
 	/* task completed - clean memory*/
@@ -211,13 +219,11 @@ cb_done:
 
 	dbus_message_unref(reply);
 	dbus_pending_call_unref(pending);
-
-	return;
 }
 
 static int bt_create_paired_device(DBusConnection *conn,
-				struct near_oob_data *oob,
-				const char *capabilities)
+						struct near_oob_data *oob,
+						const char *capabilities)
 {
 	const char *agent_path = OOB_AGENT;
 
@@ -229,7 +235,6 @@ static int bt_create_paired_device(DBusConnection *conn,
 			DBUS_TYPE_OBJECT_PATH, &agent_path,
 			DBUS_TYPE_STRING, &capabilities,
 			DBUS_TYPE_INVALID);
-
 }
 
 static void bt_oob_add_remote_data_cb(DBusPendingCall *pending, void *user_data)
@@ -267,12 +272,10 @@ cb_fail:
 
 	dbus_message_unref(reply);
 	dbus_pending_call_unref(pending);
-
-	return;
 }
 
 static int bt_oob_add_remote_data(DBusConnection *conn,
-				struct near_oob_data *oob)
+						struct near_oob_data *oob)
 {
 	int16_t hash_len = 16;
 	int16_t rdm_len = 16;
@@ -311,8 +314,7 @@ static int bt_do_pairing(struct near_oob_data *oob)
 
 /*
  */
-static int extract_properties(DBusMessage *reply,
-		struct near_oob_data *oob)
+static int extract_properties(DBusMessage *reply, struct near_oob_data *oob)
 {
 	char *data = NULL;
 	int idata;
@@ -415,6 +417,66 @@ fail:
 	return -ENOMEM;
 }
 
+static gboolean bt_adapter_property_changed(DBusConnection *conn,
+							DBusMessage *message,
+							void *user_data)
+{
+	DBusMessageIter iter;
+	DBusMessageIter var;
+	const char *property;
+
+	if (dbus_message_iter_init(message, &iter) == FALSE)
+		return TRUE;
+
+	dbus_message_iter_get_basic(&iter, &property);
+	dbus_message_iter_next(&iter);
+
+	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_VARIANT)
+		return TRUE;
+
+	dbus_message_iter_recurse(&iter, &var);
+
+	if (g_str_equal(property, "Name") == TRUE) {
+		const char *name;
+
+		if (dbus_message_iter_get_arg_type(&var) != DBUS_TYPE_STRING)
+			return TRUE;
+
+		dbus_message_iter_get_basic(&iter, &name);
+
+		g_free(bt_def_oob_data.bt_name);
+		bt_def_oob_data.bt_name = g_strdup(name);
+
+		if (bt_def_oob_data.bt_name != NULL)
+			bt_def_oob_data.bt_name_len = strlen(name);
+		else
+			bt_def_oob_data.bt_name_len = 0;
+
+		DBG("%s: %s", property, name);
+	} else if (g_str_equal(property, "Class") == TRUE) {
+		int class;
+
+		if (dbus_message_iter_get_arg_type(&var) != DBUS_TYPE_UINT32)
+			return TRUE;
+
+		dbus_message_iter_get_basic(&var, &class);
+		bt_def_oob_data.class_of_device = class;
+
+		DBG("%s: %x", property, bt_def_oob_data.class_of_device);
+	} else if (g_str_equal(property, "Powered") == TRUE) {
+		dbus_bool_t powered;
+
+		if (dbus_message_iter_get_arg_type(&var) != DBUS_TYPE_BOOLEAN)
+			return TRUE;
+
+		dbus_message_iter_get_basic(&var, &powered);
+		bt_def_oob_data.powered = powered;
+
+		DBG("%s: %u", property, bt_def_oob_data.powered);
+	}
+
+	return TRUE;
+}
 
 /* Get default local adapter properties */
 static void bt_get_properties_cb(DBusPendingCall *pending, void *user_data)
@@ -441,6 +503,12 @@ static void bt_get_properties_cb(DBusPendingCall *pending, void *user_data)
 	else
 		DBG("Get Properties complete: %s", bt_props->def_adapter);
 
+	adapter_props_watch = g_dbus_add_signal_watch(bt_conn, NULL, NULL,
+						ADAPTER_INTF,
+						ADAPTER_PROPERTY_CHANGED,
+						bt_adapter_property_changed,
+						NULL, NULL);
+
 	/* clean */
 	dbus_message_unref(reply);
 	dbus_pending_call_unref(pending);
@@ -452,8 +520,6 @@ cb_fail:
 
 	dbus_message_unref(reply);
 	dbus_pending_call_unref(pending);
-
-	return;
 }
 
 static void bt_get_default_adapter_cb(DBusPendingCall *pending, void *user_data)
@@ -501,8 +567,6 @@ cb_fail:
 
 	dbus_message_unref(reply);
 	dbus_pending_call_unref(pending);
-
-	return;
 }
 
 static int bt_refresh_adapter_props(DBusConnection *conn, void *user_data)
@@ -518,65 +582,77 @@ static int bt_refresh_adapter_props(DBusConnection *conn, void *user_data)
 }
 
 /* Parse and fill the bluetooth oob information block */
-static void bt_parse_eir(uint8_t *ptr, uint16_t bt_oob_data_size,
-		struct near_oob_data *oob)
+static void bt_parse_eir(uint8_t *eir_data, uint16_t eir_data_len,
+				struct near_oob_data *oob, uint16_t *props)
 {
-	uint8_t	eir_code;
-	uint8_t eir_length;
 	char *tmp;
+	uint16_t len = 0;
 
-	DBG("");
+	DBG("total len: %u", eir_data_len);
 
-	while (bt_oob_data_size) {
-		eir_length = *ptr++;	/* EIR length */
-		eir_code = *ptr++;	/* EIR code */
+	while (len < eir_data_len - 1) {
+		uint8_t eir_len = eir_data[0];	/* EIR field length */
+		uint8_t eir_code;		/* EIR field type*/
+		uint8_t data_len;		/* EIR data length */
+		uint8_t *data;
 
 		/* check for early termination */
-		if (eir_length == 0) {
-			bt_oob_data_size = 0;
-			continue;
-		}
+		if (eir_len == 0)
+			break;
 
-		/* Remaining bytes */
-		bt_oob_data_size = bt_oob_data_size - eir_length - EIR_SIZE_LEN;
+		len += eir_len + 1;
 
-		/* Remove len byte from eir_length */
-		eir_length = eir_length - EIR_SIZE_LEN;
+		/* Do not continue EIR Data parsing if got incorrect length */
+		if (len > eir_data_len)
+			break;
+
+		data_len = eir_len - 1;
+
+		eir_code = eir_data[1]; /* EIR code */
+		data = &eir_data[2];
+
+		DBG("type 0x%.2X data_len %u", eir_code, data_len);
 
 		switch (eir_code) {
 		case EIR_NAME_SHORT:
 		case EIR_NAME_COMPLETE:
-			oob->bt_name = g_try_malloc0(eir_length + 1); /* eos */
+			oob->bt_name = g_try_malloc0(data_len + 1); /* eos */
 			if (oob->bt_name) {
-				oob->bt_name_len = eir_length;
-				memcpy(oob->bt_name, ptr, oob->bt_name_len);
-				oob->bt_name[eir_length] = 0;	/* end str*/
+				oob->bt_name_len = data_len;
+				memcpy(oob->bt_name, data, oob->bt_name_len);
+				oob->bt_name[data_len] = 0;	/* end str*/
 			}
 			break;
 
 		case EIR_CLASS_OF_DEVICE:
 			tmp = g_strdup_printf("%02X%02X%02X",
-					*ptr, *(ptr + 1), *(ptr + 2));
-			if (tmp != NULL)
+					*data, *(data + 1), *(data + 2));
+			if (tmp != NULL) {
 				oob->class_of_device = strtol(tmp, NULL, 16);
+				*props |= OOB_PROPS_COD;
+			}
 			g_free(tmp);
 			break;
 
 		case EIR_SP_HASH:
 			oob->spair_hash = g_try_malloc0(OOB_SP_SIZE);
-			if (oob->spair_hash)
-				memcpy(oob->spair_hash, ptr, OOB_SP_SIZE);
+			if (oob->spair_hash) {
+				memcpy(oob->spair_hash, data, OOB_SP_SIZE);
+				*props |= OOB_PROPS_SP_HASH;
+			}
 			break;
 
 		case EIR_SP_RANDOMIZER:
 			oob->spair_randomizer = g_try_malloc0(OOB_SP_SIZE);
-			if (oob->spair_randomizer)
+			if (oob->spair_randomizer) {
 				memcpy(oob->spair_randomizer,
-						ptr, OOB_SP_SIZE);
+						data, OOB_SP_SIZE);
+				*props |= OOB_PROPS_SP_RANDOM;
+			}
 			break;
 
 		case EIR_SECURITY_MGR_FLAGS:
-			oob->security_manager_oob_flags = *ptr;
+			oob->security_manager_oob_flags = *data;
 			break;
 
 		case EIR_UUID128_ALL:
@@ -585,15 +661,13 @@ static void bt_parse_eir(uint8_t *ptr, uint16_t bt_oob_data_size,
 			break;
 
 		default:	/* ignore and skip */
-			near_warn("Unknown EIR: x%02x (len: %d)", eir_code,
-								eir_length);
+			near_error("Unknown EIR x%02x (len: %d)", eir_code,
+								eir_len);
 			break;
 		}
 		/* Next eir */
-		ptr = ptr + eir_length;
+		eir_data += eir_len + 1;
 	}
-
-	return;
 }
 
 /*
@@ -601,13 +675,16 @@ static void bt_parse_eir(uint8_t *ptr, uint16_t bt_oob_data_size,
  * to determine the record data structure.
  * Some specifications are proprietary (eg. "short mode")
  * and are not fully documented.
+ * mime_properties is a bitmask and should reflect the fields found in
+ * the incoming oob.
  */
-int __near_bluetooth_parse_oob_record(uint8_t version, uint8_t *bt_data,
-		near_bool_t pair)
+int __near_bluetooth_parse_oob_record(struct carrier_data *data,
+						uint16_t *mime_properties,
+						near_bool_t pair)
 {
 	struct near_oob_data *oob;
 	uint16_t bt_oob_data_size;
-	uint8_t	*ptr = bt_data;
+	uint8_t *ptr = data->data;
 	uint8_t	marker;
 	char *tmp;
 
@@ -615,13 +692,23 @@ int __near_bluetooth_parse_oob_record(uint8_t version, uint8_t *bt_data,
 
 	oob = g_try_malloc0(sizeof(struct near_oob_data));
 
-	if (version == BT_MIME_V2_1) {
-		/* Total OOB data size (including size bytes)*/
-		bt_oob_data_size = *((uint16_t *)(bt_data));
+	if (data->type == BT_MIME_V2_1) {
+		/*
+		 * Total OOB data size (including size bytes)
+		 * Some implementations (e.g. Android 4.1) stores
+		 * the data_size in big endian but NDEF forum spec (BT Secure
+		 * Simple Pairing) requires a little endian. At the same time,
+		 * the NDEF forum NDEF spec define a payload length as single
+		 * byte (and the payload size IS the oob data size).
+		 */
+		bt_oob_data_size = near_get_le16(ptr);
+		if (bt_oob_data_size > 0xFF)	/* Big Endian */
+			bt_oob_data_size = GUINT16_FROM_BE(bt_oob_data_size);
+
 		bt_oob_data_size -= 2 ; /* remove oob datas size len */
 
 		/* First item: BD_ADDR (mandatory) */
-		ptr = &bt_data[2];
+		ptr = &data->data[2];
 		oob->bd_addr = g_strdup_printf("%02X:%02X:%02X:%02X:%02X:%02X",
 				ptr[5],	ptr[4], ptr[3], ptr[2], ptr[1], ptr[0]);
 
@@ -630,8 +717,9 @@ int __near_bluetooth_parse_oob_record(uint8_t version, uint8_t *bt_data,
 		bt_oob_data_size -= BT_ADDRESS_SIZE ;
 
 		if (bt_oob_data_size)
-			bt_parse_eir(ptr, bt_oob_data_size, oob);
-	} else if (version == BT_MIME_V2_0) {
+			bt_parse_eir(ptr, bt_oob_data_size, oob,
+							mime_properties);
+	} else if (data->type == BT_MIME_V2_0) {
 		marker = *ptr++;	/* could be '$' */
 
 		oob->bd_addr = g_strdup_printf(
@@ -676,7 +764,7 @@ int __near_bluetooth_parse_oob_record(uint8_t version, uint8_t *bt_data,
 
 	/* check and get the default adapter */
 	oob->def_adapter = g_strdup(bt_def_oob_data.def_adapter);
-	if (oob->bt_name == NULL) {
+	if (oob->def_adapter == NULL) {
 		near_error("bt_get_default_adapter failed");
 		bt_eir_free(oob);
 		return -EIO;
@@ -702,8 +790,8 @@ int __near_bluetooth_pair(void *data)
 
 /* This function is synchronous as oob datas change on each session */
 static int bt_sync_oob_readlocaldata(DBusConnection *conn, char *adapter_path,
-		char *spair_hash,
-		char *spair_randomizer)
+							char *spair_hash,
+							char *spair_randomizer)
 {
 	DBusMessage *message, *reply;
 	DBusError error;
@@ -755,12 +843,12 @@ done:
 /*
  * External API to get bt properties
  * Prepare a "real" oob datas block
+ * mime_props is a bitmask we use to add or not specific fields in the
+ * oob frame (e.g.: OOB keys)
  * */
-uint8_t *__near_bluetooth_local_get_properties(int *bt_data_len)
+struct carrier_data *__near_bluetooth_local_get_properties(uint16_t mime_props)
 {
-	uint8_t *bt_oob_block = NULL;
-	uint16_t bt_oob_block_size = 0;
-	int max_block_size;
+	struct carrier_data *data = NULL;
 	uint8_t offset;
 
 	char hash[OOB_SP_SIZE];
@@ -772,104 +860,96 @@ uint8_t *__near_bluetooth_local_get_properties(int *bt_data_len)
 		goto fail;
 	}
 
-	/* Prepare the BT block */
-	max_block_size = sizeof(uint16_t) +		/* stored oob size */
-			BT_ADDRESS_SIZE +
-			EIR_HEADER_LEN + bt_def_oob_data.bt_name_len +
-			EIR_HEADER_LEN + COD_SIZE +	/* class */
-			EIR_HEADER_LEN + OOB_SP_SIZE +	/* oob hash */
-			EIR_HEADER_LEN + OOB_SP_SIZE;	/* oob random */
+	data = g_try_malloc0(sizeof(*data));
+	if (data == NULL)
+		goto fail;
 
-
-	bt_oob_block_size = sizeof(uint16_t)	/* stored oob size */
+	data->size = sizeof(uint16_t)	/* stored oob size */
 			+ BT_ADDRESS_SIZE;	/* device address */
 
-	bt_oob_block = g_try_malloc0(max_block_size);
-	if (bt_oob_block == NULL)
-		goto fail;
 	offset = sizeof(uint16_t); /* Skip size...will be filled later */
 
 	/* Now prepare data frame */
-	memcpy(bt_oob_block + offset, bt_def_oob_data.bd_addr, BT_ADDRESS_SIZE);
+	memcpy(data->data + offset, bt_def_oob_data.bd_addr, BT_ADDRESS_SIZE);
 	offset += BT_ADDRESS_SIZE;
 
-	/* bt name */
-	if (bt_def_oob_data.bt_name != NULL) {
-		bt_oob_block_size += (bt_def_oob_data.bt_name_len +
-								EIR_HEADER_LEN);
-
-		bt_oob_block[offset++] = bt_def_oob_data.bt_name_len +
-								EIR_SIZE_LEN;
-		bt_oob_block[offset++] = EIR_NAME_COMPLETE; /* EIR data type */
-		memcpy(bt_oob_block + offset, bt_def_oob_data.bt_name,
-					bt_def_oob_data.bt_name_len);
-		offset += bt_def_oob_data.bt_name_len;
-	}
-
 	/* CoD */
-	bt_oob_block_size += COD_SIZE +  EIR_HEADER_LEN;
+	data->size += COD_SIZE +  EIR_HEADER_LEN;
 
-	bt_oob_block[offset++] = COD_SIZE + EIR_SIZE_LEN;
-	bt_oob_block[offset++] = EIR_CLASS_OF_DEVICE;
+	data->data[offset++] = COD_SIZE + EIR_SIZE_LEN;
+	data->data[offset++] = EIR_CLASS_OF_DEVICE;
 
-	memcpy(bt_oob_block + offset,
+	memcpy(data->data + offset,
 			(uint8_t *)&bt_def_oob_data.class_of_device, COD_SIZE);
 	offset += COD_SIZE;
 
-	/* The following data are generated dynamically
-	 * so we have to read the local oob data
-	 * */
-	if (bt_sync_oob_readlocaldata(bt_conn, bt_def_oob_data.def_adapter,
+	/*
+	 * The following data are generated dynamically so we have to read the
+	 * local oob data. Only add OOB pairing keys if needed.
+	 */
+	if ((mime_props & OOB_PROPS_SP) != 0 &&
+			bt_sync_oob_readlocaldata(bt_conn,
+					bt_def_oob_data.def_adapter,
 					hash, random) == OOB_SP_SIZE) {
-		bt_oob_block_size += 2 * (OOB_SP_SIZE + EIR_HEADER_LEN);
+		data->size += 2 * (OOB_SP_SIZE + EIR_HEADER_LEN);
 
 		/* OOB datas */
 		if (hash != NULL) {
-			bt_oob_block[offset++] = OOB_SP_SIZE + EIR_SIZE_LEN;
-			bt_oob_block[offset++] = EIR_SP_HASH;
-			memcpy(bt_oob_block + offset, hash, OOB_SP_SIZE);
+			data->data[offset++] = OOB_SP_SIZE + EIR_SIZE_LEN;
+			data->data[offset++] = EIR_SP_HASH;
+			memcpy(data->data + offset, hash, OOB_SP_SIZE);
 			offset += OOB_SP_SIZE;
 		}
 
 		if (random != NULL) {
-			bt_oob_block[offset++] = OOB_SP_SIZE + EIR_SIZE_LEN;
-			bt_oob_block[offset++] = EIR_SP_RANDOMIZER;
-			memcpy(bt_oob_block + offset, random, OOB_SP_SIZE);
+			data->data[offset++] = OOB_SP_SIZE + EIR_SIZE_LEN;
+			data->data[offset++] = EIR_SP_RANDOMIZER;
+			memcpy(data->data + offset, random, OOB_SP_SIZE);
 			offset += OOB_SP_SIZE;
 		}
 	}
 
-	*(uint16_t *)bt_oob_block = bt_oob_block_size ;
-	*bt_data_len = bt_oob_block_size;
+	/* bt name */
+	if (bt_def_oob_data.bt_name != NULL) {
+		int name_len;
 
-	return bt_oob_block;
+		data->size += EIR_HEADER_LEN;
+
+		if (data->size + bt_def_oob_data.bt_name_len
+				> EIR_SIZE_MAX) {
+			name_len = EIR_SIZE_MAX - data->size;
+			data->data[offset++] = name_len + EIR_SIZE_LEN;
+			/* EIR data type */
+			data->data[offset++] = EIR_NAME_COMPLETE;
+		} else {
+			name_len = bt_def_oob_data.bt_name_len;
+			data->data[offset++] = name_len + EIR_SIZE_LEN;
+			/* EIR data type */
+			data->data[offset++] = EIR_NAME_SHORT;
+		}
+
+		data->size += name_len;
+		memcpy(data->data + offset, bt_def_oob_data.bt_name, name_len);
+		offset += name_len;
+	}
+
+	data->data[0] = data->size ;
+
+	if (bt_def_oob_data.powered == TRUE)
+		data->state = CPS_ACTIVE;
+	else
+		data->state = CPS_INACTIVE;
+
+	return data;
 
 fail:
-	g_free(bt_oob_block);
+	g_free(data);
 	return NULL;
 }
 
-static void bt_connect(DBusConnection *conn, void *user_data)
-{
-
-	DBG("connection %p with %p", conn, user_data);
-	if (bt_refresh_adapter_props(conn, user_data) < 0)
-		near_error("bt_get_default_adapter failed");
-
-	return;
-}
-
-static void bt_disconnect(DBusConnection *conn, void *user_data)
-{
-	DBG("%p", conn);
-
-	__bt_eir_free(user_data);
-}
-
 /* BT adapter removed handler */
-static gboolean bt_adapter_removed(DBusConnection *conn,
-					DBusMessage *message,
-					void *user_data)
+static gboolean bt_adapter_removed(DBusConnection *conn, DBusMessage *message,
+							void *user_data)
 {
 	DBusMessageIter iter;
 	struct near_oob_data *bt_props = user_data;
@@ -880,6 +960,9 @@ static gboolean bt_adapter_removed(DBusConnection *conn,
 	if (bt_props->def_adapter == NULL)
 		return TRUE;
 
+	g_dbus_remove_watch(bt_conn, adapter_props_watch);
+	adapter_props_watch = 0;
+
 	if (dbus_message_iter_init(message, &iter) == FALSE)
 		return TRUE;
 
@@ -909,6 +992,9 @@ static gboolean bt_default_adapter_changed(DBusConnection *conn,
 	if (dbus_message_iter_init(message, &iter) == FALSE)
 		return TRUE;
 
+	g_dbus_remove_watch(bt_conn, adapter_props_watch);
+	adapter_props_watch = 0;
+
 	dbus_message_iter_get_basic(&iter, &adapter_path);
 	DBG("New default adapter [%s]", adapter_path);
 
@@ -928,45 +1014,114 @@ static void bt_dbus_disconnect_cb(DBusConnection *conn, void *user_data)
 	bt_conn = NULL;
 }
 
-static guint watch;
-static guint removed_watch;
-static guint adapter_watch;
-
-static int bt_prepare_handlers(DBusConnection *conn)
+static gboolean register_bluez(gpointer user_data)
 {
-	DBG("%p", conn);
+	DBG("");
 
-	if (conn == NULL)
-		return -EIO;
+	register_bluez_timer = 0;
 
-	watch = g_dbus_add_service_watch(conn, BLUEZ_SERVICE,
-						bt_connect,
-						bt_disconnect,
-						&bt_def_oob_data, NULL);
-
-	removed_watch = g_dbus_add_signal_watch(conn, NULL, NULL, MANAGER_INTF,
+	removed_watch = g_dbus_add_signal_watch(bt_conn, NULL, NULL,
+						MANAGER_INTF,
 						ADAPTER_REMOVED,
 						bt_adapter_removed,
 						&bt_def_oob_data, NULL);
 
 
-	adapter_watch = g_dbus_add_signal_watch(conn, NULL, NULL, MANAGER_INTF,
+	adapter_watch = g_dbus_add_signal_watch(bt_conn, NULL, NULL,
+						MANAGER_INTF,
 						DEFAULT_ADAPTER_CHANGED,
 						bt_default_adapter_changed,
 						&bt_def_oob_data, NULL);
 
-	if (watch == 0 ||  removed_watch == 0 || adapter_watch == 0) {
-		near_error("Bluez event handlers failed to register.");
-		g_dbus_remove_watch(conn, watch);
-		g_dbus_remove_watch(conn, removed_watch);
-		g_dbus_remove_watch(conn, adapter_watch);
+	if (removed_watch == 0 || adapter_watch == 0) {
+		near_error("BlueZ event handlers failed to register.");
+		g_dbus_remove_watch(bt_conn, removed_watch);
+		g_dbus_remove_watch(bt_conn, adapter_watch);
 
+		return FALSE;
+	}
+
+	if (bt_refresh_adapter_props(bt_conn, user_data) < 0)
+		near_error("Failed to get BT adapter properties");
+
+	return FALSE;
+}
+
+static void bt_connect(DBusConnection *conn, void *data)
+{
+	DBG("connection %p with %p", conn, data);
+
+	if (__near_agent_handover_registered(HO_AGENT_BT) == TRUE) {
+		DBG("Agent already registered");
+		return;
+	}
+
+	/*
+	 * BlueZ 5 will register itself as HandoverAgent, give it some time
+	 * to do it before going legacy way.
+	 */
+	register_bluez_timer = g_timeout_add_seconds(AGENT_REGISTER_TIMEOUT,
+							register_bluez, data);
+}
+
+static void bt_disconnect(DBusConnection *conn, void *user_data)
+{
+	DBG("%p", conn);
+
+	/* If timer is running no BlueZ watchers were registered yet */
+	if (register_bluez_timer > 0) {
+		g_source_remove(register_bluez_timer);
+		register_bluez_timer = 0;
+		return;
+	}
+
+	__bt_eir_free(user_data);
+
+	g_dbus_remove_watch(bt_conn, removed_watch);
+	removed_watch = 0;
+
+	g_dbus_remove_watch(bt_conn, adapter_watch);
+	adapter_watch = 0;
+
+	g_dbus_remove_watch(bt_conn, adapter_props_watch);
+	adapter_props_watch = 0;
+}
+
+static int bt_prepare_handlers(DBusConnection *conn)
+{
+	if (__near_agent_handover_registered(HO_AGENT_BT) == TRUE)
+		return 0;
+
+	watch = g_dbus_add_service_watch(bt_conn, BLUEZ_SERVICE,
+						bt_connect,
+						bt_disconnect,
+						&bt_def_oob_data, NULL);
+	if (watch == 0) {
+		near_error("BlueZ service watch handler failed to register.");
+		g_dbus_remove_watch(bt_conn, watch);
 		return -EIO;
 	}
 
 	return 0;
 }
 
+void __near_bluetooth_legacy_start(void)
+{
+	DBG("");
+
+	bt_prepare_handlers(bt_conn);
+}
+
+void __near_bluetooth_legacy_stop(void)
+{
+	DBG("");
+
+	g_dbus_remove_watch(bt_conn, watch);
+	watch = 0;
+
+	bt_disconnect(bt_conn, &bt_def_oob_data);
+}
+
 /* Bluetooth exiting function */
 void __near_bluetooth_cleanup(void)
 {
@@ -975,15 +1130,9 @@ void __near_bluetooth_cleanup(void)
 	if (bt_conn == NULL)
 		return;
 
-	g_dbus_remove_watch(bt_conn, watch);
-	g_dbus_remove_watch(bt_conn, removed_watch);
-	g_dbus_remove_watch(bt_conn, adapter_watch);
+	__near_bluetooth_legacy_stop();
 
 	dbus_connection_unref(bt_conn);
-
-	__bt_eir_free(&bt_def_oob_data);
-
-	return;
 }
 
 /*
