@@ -47,7 +47,7 @@
 #define PCAP_SNAP_LEN 0xFFFF
 #define PCAP_NETWORK 0xF5
 
-#define SNAP_LEN 1024
+#define BUFFER_LEN 0x8FF
 
 static GIOChannel *gio_channel = NULL;
 guint watch = 0;
@@ -157,23 +157,36 @@ static void pcap_file_cleanup(void)
  * 00000000: 01 01 43 20 30 70 72 6F 70 65 72 74 69 65 73 20  |..C 0properties |
  *
  */
-void sniffer_print_hexdump(FILE *file, unsigned char *data, int len,
-							char *line_prefix)
+void sniffer_print_hexdump(FILE *file, guint8 *data, guint32 len,
+			   gchar *line_prefix, gboolean print_len)
 {
-	int digits;
-	int offset;
-	int total;
-	char line[LINE_SIZE];
-	char *hexa = NULL, *human = NULL;
+	guint8 digits;
+	guint32 offset;
+	guint32 total;
+	guint32 output_len;
+	gchar line[LINE_SIZE];
+	gchar *hexa = NULL, *human = NULL;
 
-	if (len <= 0)
+	if (len == 0)
 		return;
 
 	offset = 0;
 	digits = 0;
 	total = 0;
 
-	while (total < len) {
+	if (opts.snap_len && len > opts.snap_len)
+		output_len = opts.snap_len;
+	else
+		output_len = len;
+
+	if (print_len) {
+		if (line_prefix)
+			fprintf(file, "%s", line_prefix);
+
+		fprintf(file, "Total length: %u bytes\n", len);
+	}
+
+	while (total < output_len) {
 		if (digits == 0) {
 			memset(line, ' ', HUMAN_READABLE_OFFSET);
 
@@ -203,13 +216,19 @@ void sniffer_print_hexdump(FILE *file, unsigned char *data, int len,
 		total++;
 	}
 
-	if ((len & 0xF) != 0) {
+	if ((output_len & 0xF) != 0) {
 		*hexa = ' ';
 		strcpy(human, "|");
 		if (line_prefix)
 			fprintf(file, "%s", line_prefix);
 		fprintf(file, "%s\n", line);
 	}
+
+	if (output_len != len) {
+		if (line_prefix)
+			fprintf(file, "%s", line_prefix);
+		fprintf(file, "--- truncated ---\n");
+	}
 }
 
 static gboolean gio_handler(GIOChannel *channel,
@@ -244,7 +263,7 @@ static gboolean gio_handler(GIOChannel *channel,
 	msg.msg_iov = &iov;
 	msg.msg_iovlen = 1;
 	iov.iov_base = buffer;
-	iov.iov_len = opts.snap_len;
+	iov.iov_len = BUFFER_LEN;
 
 	len = recvmsg(sock, &msg, 0);
 	if (len < 0) {
@@ -288,10 +307,7 @@ int sniffer_init(void)
 	int err;
 	int one = 1;
 
-	if (opts.snap_len < SNAP_LEN)
-		opts.snap_len = SNAP_LEN;
-
-	buffer = g_malloc(opts.snap_len);
+	buffer = g_malloc(BUFFER_LEN);
 
 	sock = socket(AF_NFC, SOCK_RAW, NFC_SOCKPROTO_LLCP);
 
